IMPLEMENTATION MODULE LC;

FROM SYSTEM IMPORT BYTE;
IMPORT Types, Lib, Lib2;
IMPORT IO;

CONST
  baseMax = 10 + ( ORD('Z') + 1 - ORD('A') );

TYPE
  TCnvDesc = RECORD
    CASE : CARDINAL OF
    | 0: v: Types.TCard;
    | 1: l, h: Types.TShtCard;
    END;
  END;


PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Lib.FatalError( s );
  HALT;
END Fatal;


PROCEDURE Len( x-: ARRAY OF BYTE ): CARDINAL;
VAR
  l: CARDINAL;
BEGIN
  l := SIZE( x );
  WHILE ( l > 1 ) & ( x[l-1] = BYTE(0) ) DO DEC( l ) END;
  RETURN l;
END Len;

PROCEDURE Len0( x-: ARRAY OF BYTE ): CARDINAL;
VAR
  l: CARDINAL;
BEGIN
  l := SIZE( x );
  WHILE ( l > 0 ) & ( x[l-1] = BYTE(0) ) DO DEC( l ) END;
  RETURN l;
END Len0;

PROCEDURE Cpy( x: ARRAY OF BYTE; VAR y: ARRAY OF BYTE ): BOOLEAN;
VAR
  l, i: CARDINAL;
BEGIN
  l := Len( x );
  IF l > SIZE( y ) THEN RETURN FALSE END;
  FOR i := 0 TO l - 1 DO y[i] := x[i] END;
  WHILE l < SIZE( y ) DO y[l] := BYTE(0); INC( l ) END;
  RETURN TRUE;
END Cpy;

PROCEDURE Char( x: Types.TShtCard ): CHAR;
BEGIN
  IF x < 10 THEN
    RETURN CHR( x + ORD('0') );
  ELSIF x < baseMax THEN
    RETURN CHR( x - 10 + ORD('A') );
  ELSE
    RETURN '?';  (* ??? *)
  END;
END Char;

PROCEDURE Invert( VAR s: ARRAY OF CHAR; l: CARDINAL );
VAR
  i, j: CARDINAL;
  c: CHAR;
BEGIN
  IF l > SIZE( s ) THEN l := SIZE( s ) END;
  IF l < 2 THEN RETURN END;
  FOR i := 0 TO l DIV 2 - 1 DO
    j := l - 1 - i;
    c := s[ i ]; s[ i ] := s[ j ]; s[ j ] := c;
  END;
END Invert;


PROCEDURE CardToStr( x-: ARRAY OF BYTE; VAR s: ARRAY OF CHAR; b: CARDINAL; VAR ok: BOOLEAN );
VAR
  l: CARDINAL;

  PROCEDURE Put( c: CHAR );
  BEGIN
    IF l >= SIZE( s ) THEN ok := FALSE; RETURN END;
    s[ l ] := c;
    INC( l );
  END Put;

VAR
  i: CARDINAL;
  y: Types.TShtCard;
BEGIN
  l := 0;
  ok := TRUE;
  IF b # 16 THEN ok := FALSE; RETURN END;
  FOR i := 0 TO SIZE( x ) - 1 DO
    y := x[ i ];
    Put( Char( y MOD 10H ) );
    Put( Char( y DIV 10H ) );
  END;
  WHILE ( l > 1 ) & ( s[l-1] = '0' ) DO DEC( l ) END;
  IF l < SIZE( s ) THEN s[ l ] := 0C END;
  Invert( s, l );
END CardToStr;


PROCEDURE Add( x-, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
VAR
  ok: BOOLEAN;
  ix, iy: CARDINAL;
  lx, ly, lz: CARDINAL;

  PROCEDURE GetX( VAR c: TCnvDesc );
  BEGIN
    c.v := 0;
    IF ix < lx THEN c.l := x[ ix ]; INC( ix ) END;
  END GetX;

  PROCEDURE GetY( VAR c: TCnvDesc );
  BEGIN
    c.v := 0;
    IF iy < ly THEN c.l := y[ iy ]; INC( iy ) END;
  END GetY;

  PROCEDURE Put( x: Types.TShtCard );
  BEGIN
    IF lz < SIZE( z ) THEN
      z[ lz ] := x;
      INC( lz );
    ELSE
      ok := FALSE;
    END;
  END Put;

  PROCEDURE Fin();
  BEGIN
    WHILE lz < SIZE( z ) DO z[ lz ] := 0C; INC( lz ) END;
  END Fin;

VAR
  l, i: CARDINAL;
  cx, cy, cz: TCnvDesc;
BEGIN
  ok := TRUE;
  ix := 0; iy := 0;
  lx := Len( x ); ly := Len( y ); lz := 0;
  l := lx;
  IF l < ly THEN l := ly END;
  cz.v := 0;
  FOR i := 0 TO l - 1 DO
    GetX( cx );
    GetY( cy );
    cz.v := cx.v + cy.v + cz.h;
    Put( cz.l );
    IF cz.h > 1 THEN Fatal( 'Add: invalid carry' ) END;
  END;
  IF cz.h > 0 THEN Put( cz.h ) END;
  Fin();
  RETURN ok;
END Add;

PROCEDURE Sub( x-, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
VAR
  ok: BOOLEAN;
  ix, iy: CARDINAL;
  lx, ly, lz: CARDINAL;

  PROCEDURE GetX( VAR c: TCnvDesc );
  BEGIN
    c.v := 0;
    IF ix < lx THEN c.l := x[ ix ]; INC( ix ) END;
  END GetX;

  PROCEDURE GetY( VAR c: TCnvDesc );
  BEGIN
    c.v := 0;
    IF iy < ly THEN c.l := y[ iy ]; INC( iy ) END;
  END GetY;

  PROCEDURE Put( x: Types.TShtCard );
  BEGIN
    IF lz < SIZE( z ) THEN
      z[ lz ] := x;
      INC( lz );
    ELSE
      ok := FALSE;
    END;
  END Put;

  PROCEDURE Fin();
  BEGIN
    WHILE lz < SIZE( z ) DO z[ lz ] := 0C; INC( lz ) END;
  END Fin;

VAR
  l, i: CARDINAL;
  cx, cy, cz: TCnvDesc;
BEGIN
  ok := TRUE;
  ix := 0; iy := 0;
  lx := Len( x ); ly := Len( y ); lz := 0;
  l := lx;
  IF l < ly THEN l := ly END;
  cz.v := 0;
  FOR i := 0 TO l - 1 DO
    GetX( cx );
    GetY( cy );
    cy.v := cy.v + cz.h;
    IF cx.v >= cy.v THEN
      cz.v := cx.v - cy.v;
      IF cz.h # 0 THEN Fatal( 'Sub: Invalid carry (# 0)' ) END;
    ELSE
      cz.v := cx.v + 200H - cy.v;
      IF cz.h # 1 THEN Fatal( 'Sub: Invalid carry (# 1)' ) END;
    END;
    Put( cz.l );
    IF cz.h > 1 THEN Fatal( 'Sub: Invalid carry (> 1)' ) END;
  END;
  IF cz.h > 0 THEN ok := FALSE END;
  Fin();
  RETURN ok;
END Sub;

PROCEDURE Cmp( x-, y-: ARRAY OF BYTE ): Types.TShtInt;
VAR
  lx, ly: CARDINAL;
  cx, cy: TSht;
BEGIN
  lx := Len( x ); ly := Len( y );
  IF lx < ly THEN RETURN -1 ELSIF lx > ly THEN RETURN 1 END;
  WHILE lx > 0 DO
    DEC( lx );
    cx := x[lx]; cy := y[lx];
    IF cx < cy THEN RETURN -1 ELSIF cx > cy THEN RETURN 1 END;
  END;
  RETURN 0;
END Cmp;


PROCEDURE UpS( VAR x: ARRAY OF BYTE; n: Types.TShtCard ): BOOLEAN;
VAR
  l: CARDINAL;
BEGIN
  IF n = 0 THEN RETURN TRUE END;
  l := Len( x );
  IF l + n > SIZE( x ) THEN RETURN FALSE END;  (* zu restriktiv? x=0, l=1, n=1 erlauben? *)
  WHILE l > 0 DO DEC( l ); x[l+n] := x[l] END;
  WHILE n > 0 DO DEC( n ); x[n] := BYTE(0) END;
  RETURN TRUE;
END UpS;

PROCEDURE Mul0( x-: ARRAY OF BYTE; y: BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
VAR
  ok: BOOLEAN;
  ix: CARDINAL;
  lx, lz: CARDINAL;

  PROCEDURE Get( VAR c: TCnvDesc );
  BEGIN
    c.v := 0;
    IF ix < lx THEN c.l := x[ ix ]; INC( ix ) END;
  END Get;

  PROCEDURE Put( x: Types.TShtCard );
  BEGIN
    IF lz < SIZE( z ) THEN
      z[ lz ] := x;
      INC( lz );
    ELSE
      ok := FALSE;
    END;
  END Put;

  PROCEDURE Fin();
  BEGIN
    WHILE lz < SIZE( z ) DO z[ lz ] := 0C; INC( lz ) END;
  END Fin;

VAR
  i: CARDINAL;
  cx, cz: TCnvDesc;
BEGIN
  ok := TRUE;
  ix := 0;
  lx := Len( x ); lz := 0;
  cz.v := 0;
  FOR i := 0 TO lx - 1 DO
    Get( cx );
    cz.v := cx.v * TSht( y ) + cz.h;
    Put( cz.l );
  END;
  Put( cz.h );
  Fin();
  RETURN ok;
END Mul0;

PROCEDURE Mul( x-, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
VAR
  lx, ly: CARDINAL;
  ix, iy: CARDINAL;
  cx, cy: TSht;
  cz: TCnvDesc;
BEGIN
  lx := Len( x ); ly := Len( y );
  IF lx + ly - 1 > SIZE( z ) THEN RETURN FALSE END;
  Lib2.Fill( z, 0 );
  FOR iy := 0 TO ly - 1 DO
    cy := TSht( y[iy] );
    cz.h := 0;
    FOR ix := 0 TO lx - 1 DO
      cx := TSht( x[ix] );
      cz.l := TSht( z[ix+iy] );
      cz.v := cz.l + cx * cy + cz.h;
      z[ix+iy] := cz.l;
    END;
    IF cz.h > 0 THEN
      IF lx + iy >= SIZE( z ) THEN RETURN FALSE END;
      z[lx+iy] := cz.h;
    END;
  END;
  RETURN TRUE;
END Mul;


PROCEDURE ShiftIn( VAR x: ARRAY OF BYTE; y: BYTE );
VAR
  l: CARDINAL;
BEGIN
  l := Len0( x );
  WHILE l > 0 DO DEC( l ); x[l+1] := x[l] END;
  x[0] := y;
END ShiftIn;

PROCEDURE Div0( x-, y-: ARRAY OF BYTE; VAR z: BYTE );
VAR
  a, b, c: Types.TCard;
  t: TCardMax;
BEGIN
  a := 0; b := 100H;
  WHILE a < b DO
    c := a + ( b - a ) DIV 2;
    IF ~ Mul( y, c, t ) THEN Fatal( 'Div0: Mul failed' ) END;
    IF Cmp( t, x ) <= 0 THEN a := c + 1 ELSE b := c END;
    (* a = b, (a-1)*y <= x < a*y *)
  END;
  z := BYTE( a - 1 );
END Div0;

PROCEDURE DivA( x-, y-: ARRAY OF BYTE; VAR z, r: ARRAY OF BYTE ): BOOLEAN;

  PROCEDURE Dec( VAR x: ARRAY OF BYTE; y-: ARRAY OF BYTE );
  BEGIN
    IF ~ Sub( x, y, x ) THEN Fatal( 'DivA: Sub failed' ) END;
  END Dec;

  PROCEDURE IncZ( i: CARDINAL );
  BEGIN
    z[i] := BYTE( TSht( z[i] ) + 1 );
  END IncZ;

VAR
  l: CARDINAL;
BEGIN
  IF Len0( y ) = 0 THEN RETURN FALSE END;
  Lib2.Fill( z, 0 );
  Lib2.Fill( r, 0 );
  l := Len( x );
  IF SIZE( r ) < l THEN RETURN FALSE END;  (* worst case remainder *)
  WHILE l > 0 DO
    DEC( l );
    ShiftIn( r, x[l] );
    WHILE Cmp( r, y ) >= 0 DO
      Dec( r, y );
      IF l >= SIZE( z ) THEN RETURN FALSE END;
      IncZ( l );
    END;
  END;
  RETURN TRUE;
END DivA;

PROCEDURE DivB( x-, y-: ARRAY OF BYTE; VAR z, r: ARRAY OF BYTE ): BOOLEAN;

  PROCEDURE Dec( VAR x: ARRAY OF BYTE; y-: ARRAY OF BYTE );
  BEGIN
    IF ~ Sub( x, y, x ) THEN Fatal( 'DivB: Sub failed' ) END;
  END Dec;

  PROCEDURE IncZ( i: CARDINAL );
  BEGIN
    z[i] := BYTE( TSht( z[i] ) + 1 );
  END IncZ;

VAR
  l: CARDINAL;
  c: Types.TShtInt;
BEGIN
  IF Len0( y ) = 0 THEN RETURN FALSE END;
  Lib2.Fill( z, 0 );
  Lib2.Fill( r, 0 );
  l := Len( x );
  IF SIZE( r ) < l THEN RETURN FALSE END;  (* worst case remainder *)
  WHILE l > 0 DO
    DEC( l );
    ShiftIn( r, x[l] );
    c := Cmp( r, y );
    IF ( c >= 0 ) & ( l >= SIZE(z) ) THEN RETURN FALSE END;
    WHILE c >= 0 DO
      IncZ( l );
      Dec( r, y );
      c := Cmp( r, y );
    END;
  END;
  RETURN TRUE;
END DivB;

PROCEDURE DivC( x-, y-: ARRAY OF BYTE; VAR z, r: ARRAY OF BYTE ): BOOLEAN;

  PROCEDURE Dec( VAR x: ARRAY OF BYTE; y-: ARRAY OF BYTE; f: BYTE );
  VAR t: TCardMax;
  BEGIN
    IF ~ Mul0( y, f, t ) THEN Fatal( 'DivC: Mul0 failed' ) END;
    IF ~ Sub( x, t, x ) THEN Fatal( 'DivC: Sub failed' ) END;
  END Dec;

VAR
  l: CARDINAL;
BEGIN
  IF Len0( y ) = 0 THEN RETURN FALSE END;
  Lib2.Fill( z, 0 );
  Lib2.Fill( r, 0 );
  l := Len( x );
  IF SIZE( r ) < l THEN RETURN FALSE END;  (* worst case remainder *)
  WHILE l > 0 DO
    DEC( l );
    ShiftIn( r, x[l] );
    IF Cmp( r, y ) >= 0 THEN
      IF l >= SIZE(z) THEN RETURN FALSE END;
      Div0( r, y, z[l] );
      Dec( r, y, TSht(z[l]) );
    END;
  END;
  RETURN TRUE;
END DivC;

PROCEDURE DivD( x-, y-: ARRAY OF BYTE; VAR z, r: ARRAY OF BYTE ): BOOLEAN;

  PROCEDURE Dec( VAR x: ARRAY OF BYTE; y-: ARRAY OF BYTE );
  BEGIN
    IF ~ Sub( x, y, x ) THEN Fatal( 'DivD: Sub failed' ) END;
  END Dec;

  PROCEDURE Inc( VAR x: BYTE );
  BEGIN
    x := BYTE( TSht(x) + 1 );
  END Inc;

VAR
  l: CARDINAL;
BEGIN
  IF Len0( y ) = 0 THEN RETURN FALSE END;
  Lib2.Fill( z, 0 );
  Lib2.Fill( r, 0 );
  l := Len( x );
  IF SIZE( r ) < l THEN RETURN FALSE END;  (* worst case remainder *)
  WHILE l > 0 DO
    DEC( l );
    ShiftIn( r, x[l] );
    WHILE Cmp( r, y ) >= 0 DO
      Dec( r, y );
      IF l >= SIZE( z ) THEN RETURN FALSE END;
      Inc( z[l] );
    END;
  END;
  RETURN TRUE;
END DivD;


PROCEDURE Div2( x: ARRAY OF BYTE; y-: ARRAY OF BYTE; VAR z, r: ARRAY OF BYTE ): BOOLEAN;

  PROCEDURE Cmp( x-: ARRAY OF BYTE; i, l: CARDINAL; y-: ARRAY OF BYTE ): INTEGER;
  VAR
    ly: CARDINAL;
    cx, cy: TSht;
  BEGIN
    ly := Len( y );
    IF l < ly THEN RETURN -1 ELSIF l > ly THEN RETURN 1 END;
    WHILE l > 0 DO
      DEC( l );
      cx := x[i+l]; cy := y[l];
      IF cx < cy THEN RETURN -1 ELSIF cx > cy THEN RETURN 1 END;
    END;
    RETURN 0;
  END Cmp;

  PROCEDURE Dec( VAR x: ARRAY OF BYTE; i, l: CARDINAL; y-: ARRAY OF BYTE );
  BEGIN
  END Dec;

  PROCEDURE Inc( VAR x: BYTE );
  BEGIN
    x := BYTE( TSht(x) + 1 );
  END Inc;

VAR
  l, i: CARDINAL;
BEGIN
  Lib2.Fill( z, 0 );
  l := Len( x ); i := 0;
  WHILE l > 0 DO
    DEC( l ); INC( i );
    WHILE Cmp( x, l, i, y ) >= 0 DO
      Dec( x, l, i, y );
      Inc( z[l] );
    END;
  END;
  RETURN TRUE;
END Div2;

END LC.
